#include "bluetoothdbus.h"
#include "sessiondbusinterface.h"
#include <QDBusMessage>

BluetoothDbus::BluetoothDbus(QObject *parent)
{
    Q_UNUSED(parent);
    qDebug() << Q_FUNC_INFO << __LINE__;
    userName = QString(qgetenv("USER").toStdString().data());
    adapterList = new QMap<QString, QMap<QString, QVariant>>();
    devicesList = new QMap<QString, QMap<QString, QVariant>>();
    pairedList = new QMap<QString, QMap<QString, QVariant>>();
    connectDBusSignals();
    InitAdaptersAttr();
    InitDefaultAdapterDevicesAttr();
    registerClient();
    new SessionDbusInterface(&defaultAdapterAttr,adapterList,pairedList,this);
    qDebug() << Q_FUNC_INFO << __LINE__;
}

BluetoothDbus::~BluetoothDbus()
{
    unregisterClient();
}

void BluetoothDbus::connectDBusSignals() {
    qDebug() << Q_FUNC_INFO << __LINE__;
    TRANSFER_DBUS_SIGNAL("showTrayWidgetUISignal", showTrayWidgetUISignal());
    TRANSFER_DBUS_SIGNAL("sendTransferFilesMesg", sendTransferFilesMesgSignal(QStringList));
    TRANSFER_DBUS_SIGNAL("sendTransferDeviceMesg", sendTransferDeviceMesgSignal(QString));

    CONNECT_SYSTEM_DBUS_SIGNAL("adapterAddSignal", adapterAddSignalSLot(QMap<QString, QVariant>));
    CONNECT_SYSTEM_DBUS_SIGNAL("adapterAttrChanged", adapterAttrChangedSLot(QString, QMap<QString, QVariant>));
    CONNECT_SYSTEM_DBUS_SIGNAL("adapterRemoveSignal", adapterRemoveSignalSLot(QString));
    CONNECT_SYSTEM_DBUS_SIGNAL("deviceAddSignal", deviceAddSLot(QMap<QString, QVariant>));
    CONNECT_SYSTEM_DBUS_SIGNAL("deviceAttrChanged", deviceAttrChangedSLot(QString, QMap<QString, QVariant>));
    CONNECT_SYSTEM_DBUS_SIGNAL("deviceRemoveSignal", devRemoveSignalSLot(QString, QMap<QString, QVariant>));
    TRANSFER_SYSTEM_DBUS_SIGNAL("ActiveConnection", activeConnectionSignal(QString,QString,QString,int,int));
    CONNECT_SYSTEM_DBUS_SIGNAL("updateClient", updateClientSLot());
    TRANSFER_SYSTEM_DBUS_SIGNAL("fileStatusChanged", statusChangedSignal(QMap<QString, QVariant>));
    CONNECT_SYSTEM_DBUS_SIGNAL("fileReceiveSignal", receiveFilesSlot(QMap<QString, QVariant>));
    CONNECT_SYSTEM_DBUS_SIGNAL("startPair", devPairSignalSLot(QMap<QString, QVariant>));

    QDBusConnection::systemBus().connect(SYSTEM_ACTIVE_USER_DBUS,
                                         SYSTEM_ACTIVE_USER_PATH,
                                         SYSTEM_ACTIVE_USER_INTERFACE,
                                         "ActiveUserChange",
                                         this,SIGNAL(activeUserChangedSignal(QString)));

    QDBusConnection::sessionBus().connect(MEDIA_BATTERY_SERVICE,
                                          MEDIA_BATTERY_PATH,
                                          MEDIA_BATTERY_INTERFACE,
                                          "batteryChanged",
                                          this,
                                          SLOT(devBatteryChangedSignalSlot(QString, int)));

    qDebug() << Q_FUNC_INFO << __LINE__;
}

bool BluetoothDbus::registerClient() {
    QMap<QString, QVariant> rcParam;
    rcParam.insert("dbusid", QVariant(QDBusConnection::systemBus().baseService()));
    rcParam.insert("username", QVariant(userName));
    rcParam.insert("type", QVariant(1));
    QDBusMessage dbusMsg = CREATE_SYSTEM_METHOD_CALL("registerClient");
    dbusMsg << rcParam;
    QDBusMessage response = SYSTEM_CALL_METHOD(dbusMsg);

    if (response.type() == QDBusMessage::ReplyMessage)
    {
        rcParam = response.arguments().takeFirst().toMap();
        envPC = rcParam.value("envPC").toInt();
        return rcParam.value("result").toBool();
    } else
        return false;
}

bool BluetoothDbus::unregisterClient() {
    QDBusMessage dbusMsg = CREATE_SYSTEM_METHOD_CALL("unregisterClient");
    dbusMsg << QDBusConnection::systemBus().baseService();
    QDBusMessage response = SYSTEM_CALL_METHOD(dbusMsg);

    if (response.type() == QDBusMessage::ReplyMessage)
    {
        return response.arguments().takeFirst().toBool();
    } else
        return false;
}

void BluetoothDbus::updateClientSLot() {
    InitAdaptersAttr();
    InitDefaultAdapterDevicesAttr();
    registerClient();
}

void BluetoothDbus::adapterAttrChanged(QString address, QMap<QString, QVariant> mapAttr) {
    QMap<QString, QVariant> insertAttr;
    qDebug() << Q_FUNC_INFO << __LINE__ << mapAttr;
    if (adapterList->contains(address))
        insertAttr = adapterList->value(address);

    if (mapAttr.contains("DefaultAdapter") && insertAttr.contains("DefaultAdapter")){
        if (mapAttr.value("DefaultAdapter").toBool() != insertAttr.value("DefaultAdapter")) {
            InitDefaultAdapterDevicesAttr();
            emit defaultAdapterChanged();
        }
    }

    for (auto key : mapAttr.keys()) {
        insertAttr.remove(key);
        insertAttr.insert(key, QVariant(mapAttr.value(key)));
    }

    if (adapterList->contains(address))
        adapterList->remove(address);
    adapterList->insert(address, insertAttr);
    if(insertAttr.contains("DefaultAdapter"))
        if (insertAttr.value("DefaultAdapter").toBool())
            defaultAdapterAttr = insertAttr;

    if(mapAttr.contains("ActiveConnection"))
    {
        emit adapterAutoConnChanged(mapAttr.value("ActiveConnection").toBool());
    }

    if (mapAttr.contains("Powered")) {
        if (insertAttr.value("DefaultAdapter").toBool())
            emit powerChangedSignal(mapAttr.value("Powered").toBool());
    }

    if (mapAttr.contains("trayShow")) {
        emit showTrayIcon(mapAttr.value("trayShow").toBool());
    }

    if(mapAttr.contains("ClearPinCode"))
        if (mapAttr.value("ClearPinCode").toBool())
            emit pairAgentCanceledSignal(QString(""));
}

void BluetoothDbus::setAdapterAttrInside(QString address, QMap<QString, QVariant> mapAttr) {
    if (adapterList->contains(address))
        adapterList->remove(address);
    adapterList->insert(address, mapAttr);
}

void BluetoothDbus::deviceAttrChanged(QString address, QMap<QString, QVariant> mapAttr) {
    QMap<QString, QVariant> insertAttr;
    insertAttr = devicesList->value(address);
    for (auto key : mapAttr.keys()) {
        insertAttr.remove(key);
        insertAttr.insert(key, QVariant(mapAttr.value(key)));
    }
    if (!mapAttr.contains("Battery")
            && (mapAttr.value("Type").toInt() == Type::Headset
            || mapAttr.value("Type").toInt() == Type::Headphones
            || mapAttr.value("Type").toInt() == Type::AudioVideo)) {
        mapAttr.insert("Battery", QVariant(getDevBatteryLevel(address)));
    }
    if (!mapAttr.contains("Battery"))
        mapAttr.insert("Battery", QVariant(-1));
    devicesList->remove(address);
    devicesList->insert(address, insertAttr);
    if (mapAttr.contains("Paired"))
        emit pairAgentCanceledSignal(address);
    if (mapAttr.contains("Connecting") && !mapAttr.value("Connecting").toBool())
        emit pairAgentCanceledSignal(address);
    if (insertAttr.value("Paired").toBool())
        pairedList->insert(address, insertAttr);
}

void BluetoothDbus::setDeviceAttrInside(QString address, QMap<QString, QVariant> mapAttr) {
    if (devicesList->keys().contains(address))
        devicesList->remove(address);
    if (!mapAttr.contains("Battery")
            && (mapAttr.value("Type").toInt() == Type::Headset
            || mapAttr.value("Type").toInt() == Type::Headphones
            || mapAttr.value("Type").toInt() == Type::AudioVideo)) {
        mapAttr.insert("Battery", QVariant(getDevBatteryLevel(address)));
    }
    if (!mapAttr.contains("Battery"))
        mapAttr.insert("Battery", QVariant(-1));
    devicesList->insert(address, mapAttr);
    if (mapAttr.contains("Paired") && mapAttr.value("Paired").toBool())
        pairedList->insert(address, mapAttr);
}

int BluetoothDbus::getDevBatteryLevel(QString address) {
    QDBusInterface iface(MEDIA_BATTERY_SERVICE,
                         MEDIA_BATTERY_PATH,
                         MEDIA_BATTERY_INTERFACE,
                         QDBusConnection::sessionBus());
    if (!iface.isValid())
        return -1;
    else {
        QDBusMessage battery_response = iface.call("getBatteryLevel", address);
        int battery_int = battery_response.arguments().takeFirst().toInt();
        if(battery_int)
            return battery_int;
        else
            return -1;
    }
}

void BluetoothDbus::getAdapterAttr(QString address) {
    QMap<QString, QVariant> mapAttr;
    QDBusMessage dbusMsg = CREATE_SYSTEM_METHOD_CALL("getAdapterAttr");
    dbusMsg << address << QString("");
    QDBusMessage response = SYSTEM_CALL_METHOD(dbusMsg);

    if (response.type() == QDBusMessage::ReplyMessage)
    {
        mapAttr = QDBusReply<QMap<QString, QVariant>>(response);
    }
    if (mapAttr.isEmpty()) {
        INFO_PRINT("Adapter" << address << "Init Error");
        return;
    }
    setAdapterAttrInside(address, mapAttr);
}

void BluetoothDbus::InitAdaptersAttr() {
    QStringList adpsAddrList;
    QDBusMessage dbusMsg = CREATE_SYSTEM_METHOD_CALL("getAllAdapterAddress");
    QDBusMessage response = SYSTEM_CALL_METHOD(dbusMsg);

    if (response.type() == QDBusMessage::ReplyMessage)
    {
        adpsAddrList = response.arguments().takeFirst().toStringList();
    }
    if (adpsAddrList.isEmpty()) {
        INFO_PRINT("Adapter Address Init Error");
        return;
    }
    for (int i = 0; i < adpsAddrList.size(); i++) {
        getAdapterAttr(adpsAddrList.at(i));
    }
    if(adapterList->size() == 1)
        defaultAdapterAttr = adapterList->values().at(0);
}

void BluetoothDbus::getDevAttr(QString address) {
    QMap<QString, QVariant> mapAttr;
    QDBusMessage dbusMsg = CREATE_SYSTEM_METHOD_CALL("getDevAttr");
    dbusMsg << address;
    QDBusMessage response = SYSTEM_CALL_METHOD(dbusMsg);

    if (response.type() == QDBusMessage::ReplyMessage)
    {
        mapAttr = QDBusReply<QMap<QString, QVariant>>(response);
    }
    if (mapAttr.isEmpty()) {
        INFO_PRINT("Device" << address << "Init Error");
        return;
    }
    setDeviceAttrInside(address, mapAttr);
}

void BluetoothDbus::InitDefaultAdapterDevicesAttr() {
    pairedList->clear();
    QStringList devsAddrList;
    QDBusMessage dbusMsg = CREATE_SYSTEM_METHOD_CALL("getDefaultAdapterPairedDev");
    QDBusMessage response = SYSTEM_CALL_METHOD(dbusMsg);
    if (response.type() == QDBusMessage::ReplyMessage)
    {
        devsAddrList = response.arguments().takeFirst().toStringList();
    }
    if (devsAddrList.isEmpty()) {
        INFO_PRINT("Adapter Address Init Error");
        return;
    }
    for (int i = 0; i < devsAddrList.size(); i++) {
        getDevAttr(devsAddrList.at(i));
    }
}

void BluetoothDbus::devBatteryChangedSignalSlot(QString address, int battery) {
    if (envPC == Environment::HUAWEI)
        return;
    if (!pairedList->keys().contains(address))
        return;
    qDebug() << "devBatteryChangedSignalSLot" << address << battery;
    QMap<QString, QVariant> mapAttr = pairedList->value(address);
    mapAttr.remove("Battery");
    mapAttr.insert("Battery", QVariant(battery));
    pairedList->remove(address);
    pairedList->insert(address, mapAttr);
    emit devAttrChanged(address, pairedList->value(address));
}

void BluetoothDbus::adapterAttrChangedSLot(QString address, QMap<QString, QVariant> adapterAttr) {
    adapterAttrChanged(address, adapterAttr);
}

void BluetoothDbus::adapterRemoveSignalSLot(QString address) {
    qDebug() << Q_FUNC_INFO << __LINE__ << address;
    adapterList->remove(address);
    if (adapterList->isEmpty())
        emit existAdapter(false);
}

void BluetoothDbus::adapterAddSignalSLot(QMap<QString, QVariant> adapterAttr) {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    QString address = adapterAttr.value("Addr").toString();
    setAdapterAttrInside(address, adapterAttr);
    emit existAdapter(true);
    emit adapterAddSignal(address);
}

void BluetoothDbus::adapterChangedSLot(QString address, QMap<QString, QVariant> adapterAttr) {
    qDebug() << Q_FUNC_INFO << __LINE__ << address;
    if (adapterAttr.value("DefaultAdapter").toBool() && defaultAdapterAttr.value("Addr").toString() != address) {
        InitDefaultAdapterDevicesAttr();
        emit adapterChangedSignal();
    } else
        setAdapterAttrInside(address, adapterAttr);
}

void BluetoothDbus::deviceAddSLot(QMap<QString, QVariant> deviceAttr) {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    setDeviceAttrInside(deviceAttr.value("Addr").toString(), deviceAttr);
}

void BluetoothDbus::deviceAttrChangedSLot(QString address, QMap<QString, QVariant> deviceAttr) {
    if (!devicesList->contains(address))
        return;
    qDebug() << Q_FUNC_INFO << deviceAttr << __LINE__ ;
    deviceAttrChanged(address, deviceAttr);
    if (deviceAttr.contains("Paired"))
        emit deviceAddSignal(pairedList->value(address));
    if (deviceAttr.contains("ConnectFailedId") && pairedList->contains(address))
    {
        if (pairedList->value(address).value("ShowName") != "")
            emit connectionErrorMsg(pairedList->value(address).value("ShowName").toString(),
                                    deviceAttr.value("ConnectFailedId").toInt());
        else
            emit connectionErrorMsg(pairedList->value(address).value("Name").toString(),
                                    deviceAttr.value("ConnectFailedId").toInt());
    }
    emit devAttrChanged(address, pairedList->value(address));
}

void BluetoothDbus::devRemoveSignalSLot(QString address, QMap<QString, QVariant> attr) {
    if (attr.value("Adapter").toString() != defaultAdapterAttr.value("Addr").toString())
        return;
    if (devicesList->contains(address))
        devicesList->remove(address);
    if (!pairedList->contains(address))
        return;
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    pairedList->remove(address);
    emit devRemoveSignal(address);
}

void BluetoothDbus::devErrorSignalSLot(QString a) {
    qDebug() << "-----------------------" << a;
}

void BluetoothDbus::devPairSignalSLot(QMap<QString, QVariant> pairAttr) {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    if (pairAttr.value("type").toInt())
        emit displayPasskeySignal(pairAttr.value("dev").toString(),
                                  pairAttr.value("name").toString(),
                                  pairAttr.value("pincode").toString());
    else
        emit requestConfirmationSignal(pairAttr.value("dev").toString(),
                                       pairAttr.value("name").toString(),
                                       pairAttr.value("pincode").toString());
}

void BluetoothDbus::receiveFilesSlot(QMap<QString, QVariant> info) {
    qDebug() << Q_FUNC_INFO << info << __LINE__ ;
    if (!pairedList->contains(info.value("dev").toString()))
        return;

    QString name = info.value("name").toString();
    QMap<QString, QVariant> device = pairedList->value(info.value("dev").toString());

    if (device.contains("ShowName") && device.value("ShowName").toString() != QString(""))
        name = device.value("ShowName").toString();

    emit receiveFilesSignal(info.value("dev").toString(),
                            name,
                            info.value("filename").toString(),
                            info.value("filetype").toString(),
                            info.value("filesize").toULongLong());
}

bool BluetoothDbus::sendFiles(QString address, QStringList filelist) {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    QDBusMessage dbusMsg = CREATE_SYSTEM_METHOD_CALL("sendFiles");
    dbusMsg << userName << address << filelist;
    QDBusMessage response = QDBusConnection::systemBus().call(dbusMsg, QDBus::NoBlock);

    if (response.type() == QDBusMessage::ReplyMessage)
    {
        int res = response.arguments().takeFirst().toInt();
        switch (res) {
        case 0:
            DEBUG_PRINT("Send Succeed");
            return 0;
        default:
            DEBUG_PRINT("Send Failed");
            return -1;
        }
    }  else {
        DEBUG_PRINT("Method Call Error");
        return -1;
    }
}

void BluetoothDbus::cancelFileTransfer(QString address, int opt) {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    QMap<QString, QVariant> sendMsg;
    sendMsg.insert("dev", QVariant(address));
    sendMsg.insert("transportType", QVariant(opt));
    QDBusMessage dbusMsg = CREATE_SYSTEM_METHOD_CALL("stopFiles");
    dbusMsg << sendMsg;
    QDBusMessage response = SYSTEM_CALL_METHOD(dbusMsg);
    if (response.type() == QDBusMessage::ReplyMessage)
    {
        int res = response.arguments().takeFirst().toInt();
        switch (res) {
        case 0:
            DEBUG_PRINT("Send Succeed");
            break;
        default:
            DEBUG_PRINT("Send Failed");
        }
    }  else {
        DEBUG_PRINT("Method Call Error");
    }
}

void BluetoothDbus::openBluetoothSettings() {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    Config::OpenBluetoothSettings();
}

void BluetoothDbus::activeConnectionReply(QString dev, bool v) {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    Config::activeConnectionReply(dev, v);
}

void BluetoothDbus::pairFuncReply(QString address, bool accept) {
    qDebug() << "pair with :" << address << " reply :" << accept;
    Config::pairFuncReply(address, accept);
}


void BluetoothDbus::replyFileReceiving(QString address, bool accept) {
    qDebug() << Q_FUNC_INFO << __LINE__ << QStandardPaths::writableLocation(QStandardPaths::HomeLocation);
    Config::replyFileReceiving(address,
                               accept,
                               QDir::homePath(),
                               QStandardPaths::writableLocation(QStandardPaths::HomeLocation));
}

void BluetoothDbus::devConnect(const QString address) {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    Config::devConnect(address);
}

void BluetoothDbus::devDisconnect(const QString address) {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    Config::devDisconnect(address);
}

void BluetoothDbus::setDefaultAdapterPower(bool power) {
    qDebug() << Q_FUNC_INFO <<defaultAdapterAttr.value("Powered").toBool() << power<< __LINE__ ;
    QMap<QString, QVariant> Attr;
    if (power == defaultAdapterAttr.value("Powered").toBool())
        return;
    qDebug() << Q_FUNC_INFO <<defaultAdapterAttr.value("Powered").toBool() << power<< __LINE__ ;
    Attr.insert("Powered", QVariant(power));
    Config::setDefaultAdapterAttr(Attr);
}

QMap<QString, QMap<QString, QVariant>> BluetoothDbus::getSendableDevices() {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    QMap<QString, QMap<QString, QVariant>> ret;
    for (auto addr : pairedList->keys()) {
        if (pairedList->value(addr).value("FileTransportSupport").toBool()) {
            QMap<QString, QVariant> inode = pairedList->value(addr);
            ret.insert(addr, inode);
        }
    }
    return ret;
}

bool BluetoothDbus::getDevSupportFileSend(QString address) {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    if (!pairedList->contains(address))
        return false;
    return pairedList->value(address).value("FileTransportSupport").toBool();;
}

bool BluetoothDbus::getTrayIconShow() {
    qDebug() << Q_FUNC_INFO << __LINE__;
    if (adapterList->isEmpty())
        return false;
    return defaultAdapterAttr.value("trayShow").toBool();
}

bool BluetoothDbus::getExistAdapter() {
    qDebug() << Q_FUNC_INFO << __LINE__;
    return !adapterList->isEmpty();
}

bool BluetoothDbus::isPowered() {
    qDebug() << Q_FUNC_INFO << defaultAdapterAttr.value("Powered").toBool() << __LINE__ ;
    return defaultAdapterAttr.value("Powered").toBool();
}

QMap<QString, QMap<QString, QVariant>> BluetoothDbus::getPairedDevices() {
    qDebug() << Q_FUNC_INFO << __LINE__ ;
    return *pairedList;
}
